# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: __version__.py
@time: 2019-02-21 15:22

"""

__version__ = '0.2.4'
